////////////////////////////////////////
//
// GraphCuts Segmentation using Blur
// writen by tomoyuki nagahashi
//
///////////////////////////////////////

#include "GraphCutsBlur.h"

GraphCutsBlur::GraphCutsBlur() 
{
}


GraphCutsBlur::~GraphCutsBlur() 
{
}

Mat GraphCutsBlur::Run(Mat &input, Mat &objseed, Mat &bkgseed, double lambda)
{
	m_Lambda = lambda;

	Mat result;
	
	ImagePyramid images;
	images.SetImage(input);
	
	ImagePyramid objseeds, bkgseeds;
	objseeds.SetSeed(objseed);	
	bkgseeds.SetSeed(bkgseed);

	// GMM 
	cout << "GMM learn" << endl;
	GMM objmodel, bkgmodel;
	objmodel.Init(3,5);
	bkgmodel.Init(3,5);
	for(int i = 0; i < input.rows*input.cols; i++)
	{
		vector<double> x(3);
		x[0] = input.data[i*3 + 0];
		x[1] = input.data[i*3 + 1];
		x[2] = input.data[i*3 + 2];
		if(objseed.data[i] > 127)
		{
			objmodel.SetData(x);
		}
		if(bkgseed.data[i] > 127)
		{
			bkgmodel.SetData(x);
		}
	}
	objmodel.Learn();
	bkgmodel.Learn();

	
	Mat objdist( input.size(), CV_64FC1, Scalar(0.0));
	Mat bkgdist( input.size(), CV_64FC1, Scalar(0.0));

	
	for(int i = images.GetLevel()-1; i >= 0; i--)
	{
		Mat resize_objdist, resize_bkgdist;
		resize(objdist, resize_objdist, images.GetImage(i, 0).size(), 0, 0, INTER_LINEAR);
		resize(bkgdist, resize_bkgdist, images.GetImage(i, 0).size(), 0, 0, INTER_LINEAR);
		
		Mat resize_objseed = objseeds.GetImage(i,0);
		Mat resize_bkgseed = bkgseeds.GetImage(i,0);

		
		for(int j = images.GetStep(i)-1; j >= 0; j--)
		{
			cout << "segmenetaion, level:" << i << " sigma:" << j << endl;
			Mat source = images.GetImage(i,j);
			// graph cuts
			Mat segment = Segmentaiton(
								source, 
								resize_objseed, resize_bkgseed,
								objmodel, bkgmodel,
								resize_objdist, resize_bkgdist
							);
			
			result = segment.clone();
			
			// distance transform
			cout << "distance transform" << endl;
			DistanceTransform(segment, resize_objdist, resize_bkgdist);
						
			// GMM learn
			cout << "GMM learn" << endl;
			objmodel.ClearData();
			bkgmodel.ClearData();
			for(int k = 0; k < input.rows*input.cols; k++)
			{
				vector<double> x(3);
				x[0] = input.data[k*3 + 0];
				x[1] = input.data[k*3 + 1];
				x[2] = input.data[k*3 + 2];
				if(objseed.data[k] > 127)
				{
					objmodel.SetData(x);
				}
				if(bkgseed.data[k] > 127)
				{
					bkgmodel.SetData(x);
				}
			}
			for(int k = 0; k < segment.rows*segment.cols; k++)
			{
				vector<double> x(3);
				x[0] = source.data[k*3+0];
				x[1] = source.data[k*3+1];
				x[2] = source.data[k*3+2];
				if(segment.data[k] > 127)
				{
					objmodel.SetData(x);
				}
				else
				{
					bkgmodel.SetData(x);
				}
			}
			objmodel.Learn();
			bkgmodel.Learn();
		}
		
		resize(resize_objdist, objdist, objdist.size(), 0, 0, INTER_LINEAR);
		resize(resize_bkgdist, bkgdist, bkgdist.size(), 0, 0, INTER_LINEAR);
		
	}
	
	
	
	return result;
}



//////////////////////////////////
// Graph Cuts Segmentaiton
//////////////////////////////////
Mat GraphCutsBlur::Segmentaiton(Mat &input, 
			Mat &objseed, Mat &bkgseed, 
			GMM &objmodel, GMM &bkgmodel,
			Mat &objdist, Mat &bkgdist
	)
{
	int width = input.cols;
	int height = input.rows;
	
	typedef Graph<double, double, double> GraphType;
	GraphType graph(width*height, width*height*8);
	
	graph.add_node(width*height);
	
	// n-links
	for(int i = 0; i < height; i++)
	{
		for(int j = 0; j < width; j++)
		{
			for(int dy = -1; dy <= 1; dy++)
			{
				for(int dx = -1; dx <= 1; dx++)
				{
					if(dx == 0 && dy == 0){
						continue;
					}
					if(i+dy >= 0 && i+dy < height &&
					   j+dx >= 0 && j+dx < width)
					{
						double val = DistanceColor(input, j, i, j+dx, i+dy);
						graph.add_edge( i*width+j, (i+dy)*width+(j+dx), val, 0);
					}
				}
			}
		}
	}
	
	// t-links
	for(int i = 0; i < height; i++)
	{
		for(int j = 0; j < width; j++)
		{
			vector<double> x(3);
			x[0] = input.data[i*width*3 + j*3 + 0];
			x[1] = input.data[i*width*3 + j*3 + 1];
			x[2] = input.data[i*width*3 + j*3 + 2];
			
			double sink   = -log( objmodel.Prob(x) * objdist.at<double>(i,j) ) * m_Lambda;
			double source = -log( bkgmodel.Prob(x) * bkgdist.at<double>(i,j) ) * m_Lambda;			
			
			if(objseed.data[i*width+j] > 0)
			{
				sink = 0;
				source = 1.0e+15;
			}
			if(bkgseed.data[i*width+j] > 0)
			{
				sink = 1.0e+15;
				source= 0;
			}
			graph.add_tweights(i*width+j, source, sink);
		}
	}
	
	// maxflow
	cout << "graph cuts" << endl;
	graph.maxflow();
	cout << "end" << endl;
	
	Mat result(height, width, CV_8UC1);
	for(int i = 0; i < height; i++)
	{
		for(int j = 0; j < width; j++)
		{
			if(graph.what_segment(i*width+j) == GraphType::SOURCE)
			{	
				result.data[i*width + j] = 255;
			}
			else
			{
				result.data[i*width + j] = 0;
			}
		}
	}
	
	return result;
}


/////////////////////////////
// color feature
////////////////////////////
double GraphCutsBlur::DistanceColor(Mat &input, int x1, int y1, int x2, int y2)
{
	double diff_B = input.data[y1*input.cols*3+x1*3+0] - input.data[y2*input.cols*3+x2*3+0];
	double diff_G = input.data[y1*input.cols*3+x1*3+1] - input.data[y2*input.cols*3+x2*3+1];
	double diff_R = input.data[y1*input.cols*3+x1*3+2] - input.data[y2*input.cols*3+x2*3+2];

	double d = sqrt(diff_B*diff_B + diff_G*diff_G + diff_R*diff_R);
	double dist = sqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2) * 1.0);
	
	return exp(-d/10.0) / dist;
}

//////////////////////////
// distance transform
//////////////////////////
int GraphCutsBlur::DistanceTransform(Mat &input, Mat &objdist, Mat &bkgdist)
{
	Mat input_obj = input.clone();
	Mat input_bkg = input.clone();
	
	for(int i = 0; i < input.rows*input.cols; i++)
	{
		if(input.data[i] > 127)
		{
			input_obj.data[i] = 0;
			input_bkg.data[i] = 255;
		}
		else
		{
			input_obj.data[i] = 255;
			input_bkg.data[i] = 0;
		}
	}
	
	Mat obj, bkg;
	distanceTransform(input_obj, obj, CV_DIST_L2, 3);
	distanceTransform(input_bkg, bkg, CV_DIST_L2, 3);
	
	double max = 15;
	for(int i = 0; i < objdist.rows; i++)
	{
		for(int j = 0; j < objdist.cols; j++)
		{
			if( obj.at<float>(i,j) < bkg.at<float>(i,j) )
			{
				double d = bkg.at<float>(i,j)/max * 0.5;
				objdist.at<double>(i,j) = truncate<double>(d, 0, 0.5) + 0.5;
				bkgdist.at<double>(i,j) = 1.0 - objdist.at<double>(i,j);
			}
			else
			{
				double d = obj.at<float>(i,j)/max * 0.5;
				bkgdist.at<double>(i,j) = truncate<double>(d, 0, 0.5) + 0.5;;
				objdist.at<double>(i,j) = 1.0 - bkgdist.at<double>(i,j);
			}
		}
	}
	
	
	return 0;
}

//////////////////
// truncate
/////////////////
template<typename T> T GraphCutsBlur::truncate(T a, T min, T max)
{
	if(a<min)
	{
		return min;
	}
	if(a>max)
	{
		return max;
	}
	return a;
}