////////////////////////////////////
//
// GraphCuts Segmentation
// writen by tomoyuki nagahashi
//
////////////////////////////////////

#include "GraphCutsSegmentation.h"

GraphCutsSegmentation::GraphCutsSegmentation() {
}

GraphCutsSegmentation::~GraphCutsSegmentation() {
}

Mat GraphCutsSegmentation::Run(Mat &input, Mat &objseed, Mat &bkgseed, double lambda)
{
	int width = input.cols;
	int height = input.rows;
	
	GMM objmodel, bkgmodel;
	objmodel.Init(3, 5);
	bkgmodel.Init(3, 5);
	
	cout << "set seed" << endl;
	for(int i = 0; i < height; i++)
	{
		for(int j = 0; j < width; j++)
		{
			vector<double> x(3);
			x[0] = input.data[i*width*3 + j*3 + 0];
			x[1] = input.data[i*width*3 + j*3 + 1];
			x[2] = input.data[i*width*3 + j*3 + 2];
			if(objseed.data[i*width+j] > 127)
			{
				objmodel.SetData(x);
			}
			if(bkgseed.data[i*width+j] > 127)
			{
				bkgmodel.SetData(x);
			}
		}	
	}
	
	// GMM fitting
	cout << "Fitting GMM" << endl;
	objmodel.Learn();
	bkgmodel.Learn();
	
	typedef Graph<double, double, double> GraphType;
	GraphType graph(width*height, width*height*8);
	
	graph.add_node(width*height);
	
	// n-links
	for(int i = 0; i < height; i++)
	{
		for(int j = 0; j < width; j++)
		{
			for(int dy = -1; dy <= 1; dy++)
			{
				for(int dx = -1; dx <= 1; dx++)
				{
					if(dx == 0 && dy == 0){
						continue;
					}
					if(i+dy >= 0 && i+dy < height &&
					   j+dx >= 0 && j+dx < width)
					{
						double val = DistanceColor(input, j, i, j+dx, i+dy);
						graph.add_edge( i*width+j, (i+dy)*width+(j+dx), val, 0);
					}
				}
			}
		}
	}
	
	// t-links
	for(int i = 0; i < height; i++)
	{
		for(int j = 0; j < width; j++)
		{
			vector<double> x(3);
			x[0] = input.data[i*width*3 + j*3 + 0];
			x[1] = input.data[i*width*3 + j*3 + 1];
			x[2] = input.data[i*width*3 + j*3 + 2];
			
			double sink   = -log( objmodel.Prob(x) ) * lambda;
			double source = -log( bkgmodel.Prob(x) ) * lambda;
			
			//double p_obj = objmodel.Prob(x);
			//double p_bkg = bkgmodel.Prob(x);
			//double sink   = p_obj/(p_obj+p_bkg) * lambda;
			//double source = p_bkg/(p_obj+p_bkg) * lambda;
			
			
			if(objseed.data[i*width+j] > 0)
			{
				sink = 0;
				source = 1.0e+15;
			}
			if(bkgseed.data[i*width+j] > 0)
			{
				sink = 1.0e+15;
				source= 0;
			}
			graph.add_tweights(i*width+j, source, sink);
		}
	}
	
	// maxflow
	cout << "graph cuts" << endl;
	graph.maxflow();
	cout << "end" << endl;
	
	Mat result(height, width, CV_8UC1);
	for(int i = 0; i < height; i++)
	{
		for(int j = 0; j < width; j++)
		{
			if(graph.what_segment(i*width+j) == GraphType::SOURCE)
			{	
				result.data[i*width + j] = 255;
			}
			else
			{
				result.data[i*width + j] = 0;
			}
		}
	}
	
	return result;
}


/////////////////////////////
// between color
////////////////////////////
double GraphCutsSegmentation::DistanceColor(Mat &input, int x1, int y1, int x2, int y2)
{
	double diff_B = input.data[y1*input.cols*3+x1*3+0] - input.data[y2*input.cols*3+x2*3+0];
	double diff_G = input.data[y1*input.cols*3+x1*3+1] - input.data[y2*input.cols*3+x2*3+1];
	double diff_R = input.data[y1*input.cols*3+x1*3+2] - input.data[y2*input.cols*3+x2*3+2];

	double d = sqrt(diff_B*diff_B + diff_G*diff_G + diff_R*diff_R);
	double dist = sqrt((x1-x2)*(x1-x2) + (y1-y2)*(y1-y2) * 1.0);
	
	return exp(-d/10.0) / dist;
}


