////////////////////////////////////
//
// GraphCuts Segmentation
// written by tomoyuki nagahashi
//
////////////////////////////////////

#include <stdio.h>
#include <iostream>
#include <opencv2/opencv.hpp>
#include "GraphCutsBlur.h"
#include "GraphCutsSegmentation.h"

#ifdef _DEBUG

#pragma comment(lib,"C:\\OpenCV2.2\\lib\\opencv_core220d.lib")            // opencv_core
#pragma comment(lib,"C:\\OpenCV2.2\\lib\\opencv_imgproc220d.lib")        // opencv_imgproc
#pragma comment(lib,"C:\\OpenCV2.2\\lib\\opencv_highgui220d.lib")        // opencv_highgui
#pragma comment(lib,"C:\\OpenCV2.2\\lib\\opencv_objdetect220d.lib")    // opencv_objdetect

#else

#pragma comment(lib,"C:\\OpenCV-2.2.0\\build\\lib\\Release\\opencv_core220.lib")            // opencv_core
#pragma comment(lib,"C:\\OpenCV-2.2.0\\build\\lib\\Release\\opencv_imgproc220.lib")        // opencv_imgproc
#pragma comment(lib,"C:\\OpenCV-2.2.0\\build\\lib\\Release\\opencv_highgui220.lib")        // opencv_highgui
#pragma comment(lib,"C:\\OpenCV-2.2.0\\build\\lib\\Release\\opencv_objdetect220.lib")    // opencv_objdetect

#endif

using namespace std;
using namespace cv;

Mat input;
Mat result_mask;
Mat result_mask_blur;
Mat objseed;
Mat bkgseed;
bool seedswitch = true;
bool mousedrag = false;
int old_x, old_y;

///////////////////
// mouse event
///////////////////
void mouseEvent(int event, int x, int y ,int flags, void* param)
{
	int obj, bkg;
	if(seedswitch)
	{
		obj = 255;
		bkg = 0;
	}
	else
	{
		obj = 0;
		bkg = 255;
	}
	
	switch(event)
    {
		case CV_EVENT_LBUTTONDOWN:
			mousedrag = true;
			old_x = x;
			old_y = y;
			break;
		case CV_EVENT_MOUSEMOVE:
			if(mousedrag)
			{
				for(int i = 0; i < 100; i++)
				{
					double a = i/100.0;
					int yy = y*a + (1-a)*old_y;
					int xx = x*a + (1-a)*old_x;
					objseed.data[(yy)*objseed.cols+(xx)] = obj;
					bkgseed.data[(yy)*bkgseed.cols+(xx)] = bkg;
					
					objseed.data[(yy+1)*objseed.cols+(xx)] = obj;
					bkgseed.data[(yy+1)*bkgseed.cols+(xx)] = bkg;
					objseed.data[(yy-1)*objseed.cols+(xx)] = obj;
					bkgseed.data[(yy-1)*bkgseed.cols+(xx)] = bkg;
					objseed.data[(yy)*objseed.cols+(xx+1)] = obj;
					bkgseed.data[(yy)*bkgseed.cols+(xx+1)] = bkg;
					objseed.data[(yy)*objseed.cols+(xx-1)] = obj;
					bkgseed.data[(yy)*bkgseed.cols+(xx-1)] = bkg;
				}
				old_x = x;
				old_y = y;
				
			}
			break;
		case CV_EVENT_LBUTTONUP:
			mousedrag = false;
			break;
	}
}

////////////////////
// draw image
////////////////////
int drawImage(void)
{
	Mat draw = input.clone();
	Mat result1 = input.clone();
	Mat result2 = input.clone();
	for(int i = 0; i < draw.rows; i++)
	{
		for(int j = 0; j < draw.cols; j++)
		{
			if(objseed.data[i*objseed.cols + j] == 255)
			{
				draw.data[i*draw.cols*3 + j*3 + 0] = 255;
				draw.data[i*draw.cols*3 + j*3 + 1] = 0;
				draw.data[i*draw.cols*3 + j*3 + 2] = 0;
			}
			if(bkgseed.data[i*objseed.cols + j] == 255)
			{
				draw.data[i*draw.cols*3 + j*3 + 0] = 0;
				draw.data[i*draw.cols*3 + j*3 + 1] = 0;
				draw.data[i*draw.cols*3 + j*3 + 2] = 255;
			}
			
			if(result_mask.data[i*objseed.cols + j] == 0)
			{
				result1.data[i*draw.cols*3 + j*3 + 0] /= 10;
				result1.data[i*draw.cols*3 + j*3 + 1] /= 10;
				result1.data[i*draw.cols*3 + j*3 + 2] /= 10;
			}
			
			if(result_mask_blur.data[i*objseed.cols + j] == 0)
			{
				result2.data[i*draw.cols*3 + j*3 + 0] /= 10;
				result2.data[i*draw.cols*3 + j*3 + 1] /= 10;
				result2.data[i*draw.cols*3 + j*3 + 2] /= 10;
			}
		}
	}
	
	imshow("src", draw);
	imshow("result", result1);
	imshow("result_blur", result2);
	
	return 0;
}



///////////////////
// main 
///////////////////
int main(int argc, char** argv) 
{
	
	if(argc != 2)
	{
		input = imread("./INPUT_IMAGE/banana.jpg", 1);
	}
	else
	{
		input = imread(argv[1], 1);
	}
	
	cout << "Load Image" << endl;
	int cols = input.cols;
	int rows = input.rows;
	objseed.create(rows, cols, CV_8UC1);
	bkgseed.create(rows, cols, CV_8UC1);
	result_mask.create(rows, cols, CV_8UC1);
	result_mask_blur.create(rows, cols, CV_8UC1);



	// create window
	namedWindow("result");
	namedWindow("result_blur");
	namedWindow("src");	
	
	
	// muse event func		
	cvSetMouseCallback("src", mouseEvent);
	
	GraphCutsSegmentation gc;
	GraphCutsBlur gcblur;
	
	while(1)
	{
		int key = waitKey(100);
		switch(key)
		{
			case 'q':
				return 0;
			case 's':
				if(seedswitch)
					seedswitch = false;
				else
					seedswitch = true;
				cout << seedswitch?"obj seed":"bkg seed";
				break;
			case 'r':
				result_mask_blur = gcblur.Run(input, objseed, bkgseed, 0.001);
				result_mask  = gc.Run(input, objseed, bkgseed, 0.001);
				break;
		}
		
		drawImage();
	}
	
	return 0;
}

